"""Static action sets for binary to discrete action space wrappers."""


# actions for the simple run right environment
RIGHT_ONLY = [
    ['NOP'],
    ['right'],
    ['right', 'A'],
    ['right', 'B'],
    ['right', 'A', 'B'],
]


# actions for very simple movement
SIMPLE_MOVEMENT = [
    ['NOP'],
    ['right'],
    ['right', 'A'],
    ['right', 'B'],
    ['right', 'A', 'B'],
    ['A'],
    ['left'],
]


# actions for more complex movement
COMPLEX_MOVEMENT = [
    ['NOP'],
    ['right'],
    ['right', 'A'],
    ['right', 'B'],
    ['right', 'A', 'B'],
    ['A'],
    ['left'],
    ['left', 'A'],
    ['left', 'B'],
    ['left', 'A', 'B'],
    ['down'],
]



"""An enumeration of valid ROM hacks to load into the emulator."""
import enum


@enum.unique
class RomMode(enum.Enum):
    """An enumeration of valid ROM hacks to load into the emulator."""
    # the standard ROM with no modifications
    VANILLA = 0
    # down-sampled ROM with static artifacts removed
    DOWNSAMPLE = 1
    # a simpler pixelated version of graphics
    PIXEL = 2
    # an even simpler rectangular version of graphics
    RECTANGLE = 3


# explicitly define the outward facing API of this module
__all__ = [RomMode.__name__]



"""An OpenAI Gym environment for Super Mario Bros. and Lost Levels (levels)."""
from .smb_env import SuperMarioBrosEnv
from ._rom_mode import RomMode


class SuperMarioBrosLevelEnv(SuperMarioBrosEnv):
    """An environment for playing Super Mario Bros Levels with OpenAI Gym."""

    def __init__(self,
        frameskip=1,
        max_episode_steps=float('inf'),
        rom_mode=RomMode.VANILLA,
        lost_levels=False,
        target_world=1,
        target_level=1,
    ):
        """
        Initialize a new Super Mario Bros environment.

        Args:
            frameskip (int): the number of frames to skip between steps
            max_episode_steps (float): number of steps before an episode ends
            rom_mode (RomMode): the ROM mode to use when loading ROMs from disk
            lost_levels (bool): whether to load the ROM with lost levels.
                - False: load original Super Mario Bros.
                - True: load Super Mario Bros. Lost Levels
            target_world (int): the world to target in the ROM
            target_level (int): the level to target in the given world

        Returns:
            None

        """
        # Type and value check the target world parameter
        if not isinstance(target_world, int):
            raise TypeError('target_world must be of type: int')
        if lost_levels:
            if target_world > 12:
                worlds = set(range(1, 12 + 1))
                raise ValueError('target_world')
        elif target_world > 8:
            worlds = set(range(1, 8 + 1))
            raise ValueError('target_world must be in '.format(worlds))

        # Type and value check the target level parameter
        if not isinstance(target_level, int):
            raise TypeError('target_level must be of type: int')
        if target_level > 4:
            levels = set(range(1, 4 + 1))
            raise ValueError('target_level must be in '.format(levels))

        # setup target area if target world and level are specified
        target_area = target_level
        # setup the target area depending on whether this is SMB 1 or 2
        if lost_levels:
            # setup the target area depending on the target world and level
            if target_world in {1, 3}:
                if target_level >= 2:
                    target_area = target_area + 1
            elif target_world >= 5:
                # TODO: figure out why all worlds greater than 5 fail.
                # target_area = target_area + 1
                # for now just raise a value error
                worlds = set(range(5, 12 + 1))
                msg = 'lost levels worlds {} not supported'.format(worlds)
                raise ValueError(msg)
        else:
            # setup the target area depending on the target world and level
            if target_world in {1, 2, 4, 7}:
                if target_level >= 2:
                    target_area = target_area + 1

        # copy the values to local instance variables
        self._target_world = target_world
        self._target_level = target_level
        self._target_area = target_area

        # initialize the super object
        super(SuperMarioBrosLevelEnv, self).__init__(
            frameskip=frameskip,
            max_episode_steps=max_episode_steps,
            rom_mode=rom_mode,
            lost_levels=lost_levels,
        )

    # MARK: RAM Hacks

    def _write_level(self):
        """Write the level data to RAM to overwrite the loading level."""
        self._write_mem(0x075f, self._target_world - 1)
        self._write_mem(0x075c, self._target_level - 1)
        self._write_mem(0x0760, self._target_area - 1)

    def _skip_start_screen(self):
        # press and release the start button
        self._frame_advance(8)
        self._frame_advance(0)
        # Press start until the game starts
        while self._get_time() == 0:
            # press the start button
            self._frame_advance(8)
            # force overwrite the level that is set to load
            self._write_level()
            # release the start button
            self._frame_advance(0)
            # run-out the prelevel timer to skip the animation
            self._runout_prelevel_timer()
        # after the start screen idle to skip some extra frames
        while self._get_time() >= self._time_left:
            self._time_left = self._get_time()
            self._frame_advance(8)
            self._frame_advance(0)

    def _get_done(self):
        """Return True if the episode is over, False otherwise."""
        return (
            self._get_is_dying() or
            self._get_is_dead() or
            self._get_is_world_over() or
            self._get_is_level_over()
        )


# explicitly define the outward facing API of this module
__all__ = [SuperMarioBrosLevelEnv.__name__]



"""Registration code of Gym environments in this package."""
from .smb_env import SuperMarioBrosEnv
from .smb_level_env import SuperMarioBrosLevelEnv
from ._registration import make


# define the outward facing API of this package
__all__ = [
    make.__name__,
    SuperMarioBrosEnv.__name__,
    SuperMarioBrosLevelEnv.__name__,
]



"""Super Mario Bros for OpenAI Gym."""
import os
import argparse
from nes_py.wrappers import wrap
from nes_py.wrappers import BinarySpaceToDiscreteSpaceEnv
from nes_py._app.play import play_human, play_random
from .._registration import make
from ..actions import RIGHT_ONLY, SIMPLE_MOVEMENT, COMPLEX_MOVEMENT


# The play modes for the system
_PLAY_MODES = {
    'human': play_human,
    'random': play_random
}


# a key mapping of action spaces to wrap with
_ACTION_SPACES = {
    'right': RIGHT_ONLY,
    'simple': SIMPLE_MOVEMENT,
    'complex': COMPLEX_MOVEMENT,
}


def _get_args():
    """Parse command line arguments and return them."""
    parser = argparse.ArgumentParser(description=__doc__)
    # add the argument for the Super Mario Bros environment to run
    parser.add_argument('--env', '-e',
        type=str,
        default='SuperMarioBros-v0',
        help='The name of the environment to play'
    )
    # add the argument for the mode of execution as either human or random
    parser.add_argument('--mode', '-m',
        type=str,
        default='human',
        choices=['human', 'random'],
        help='The execution mode for the emulation'
    )
    # add a flag for wrapping the environment
    parser.add_argument('--wrap', '-W',
        action='store_true',
        help='A flag to use the standard wrap while playing'
    )
    # add the argument for adjusting the action space
    parser.add_argument('--actionspace', '-a',
        type=str,
        default='nes',
        choices=['nes', 'right', 'simple', 'complex'],
        help='the action space wrapper to use'
    )
    # parse arguments and return them
    return parser.parse_args()


def main():
    """The main entry point for the command line interface."""
    # parse arguments from the command line (argparse validates arguments)
    args = _get_args()
    # build the environment with the given ID
    env = make(args.env)
    # wrap the environment with an action space if specified
    if args.actionspace != 'nes':
        print(args.actionspace)
        # unwrap the actions list by key
        actions = _ACTION_SPACES[args.actionspace]
        # wrap the environment with the new action space
        env = BinarySpaceToDiscreteSpaceEnv(env, actions)
    # wrap the environment if specified
    if args.wrap:
        env = wrap(env, agent_history_length=None)
    # play the environment with the given mode
    _PLAY_MODES[args.mode](env)


# explicitly define the outward facing API of this module
__all__ = [main.__name__]



"""The setup script for installing the package."""
from setuptools import setup, find_packages


def README():
    """Return the contents of the read me file for this project."""
    with open('README.md') as readme:
        return readme.read()


setup(
    name='gym_super_mario_bros',
    version='3.1.1',
    description='Super Mario Bros. for OpenAI Gym',
    long_description=README(),
    long_description_content_type='text/markdown',
    keywords=' '.join([
        'OpenAI-Gym',
        'NES',
        'Super-Mario-Bros',
        'Lost-Levels',
        'Reinforcement-Learning-Environment',
    ]),
    classifiers=[
        'License :: Free For Educational Use',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    url='https://github.com/Kautenja/gym-super-mario-bros',
    author='Christian Kauten',
    author_email='kautencreations@gmail.com',
    license='Proprietary',
    packages=find_packages(),
    package_data={
        'gym_super_mario_bros': ['lua/*.lua', 'roms/*.nes']
    },
    install_requires=[
        'gym>=0.10.5',
        'matplotlib>=2.0.2',
        'nes-py>=1.0.0',
        'numpy>=1.14.2',
        'opencv-python>=3.4.0.12',
        'pygame>=1.9.3',
        'pyglet>=1.3.2',
        'tqdm>=4.19.5',
    ],
    entry_points={
        'console_scripts': [
            'gym_super_mario_bros = gym_super_mario_bros._app.cli:main',
        ],
    },
)
